<?php

namespace SuperRhino\AppStoreServerApi;

use Firebase\JWT\JWT;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\GuzzleException;
use SuperRhino\AppStoreServerApi\Request\ConsumptionRequest;
use SuperRhino\AppStoreServerApi\Request\ExtendRenewalDateRequest;
use SuperRhino\AppStoreServerApi\Request\NotificationHistoryRequest;
use SuperRhino\AppStoreServerApi\Request\TransactionHistoryRequest;
use SuperRhino\AppStoreServerApi\Response\CheckTestNotificationResponse;
use SuperRhino\AppStoreServerApi\Response\ExtendRenewalDateResponse;
use SuperRhino\AppStoreServerApi\Response\HistoryResponse;
use SuperRhino\AppStoreServerApi\Response\NotificationHistoryResponse;
use SuperRhino\AppStoreServerApi\Response\OrderLookupResponse;
use SuperRhino\AppStoreServerApi\Response\RefundLookupResponse;
use SuperRhino\AppStoreServerApi\Response\SendTestNotificationResponse;
use SuperRhino\AppStoreServerApi\Response\StatusResponse;
use SuperRhino\Utils\GuzzleClient\GuzzleClient;

class AppStoreServerApi {

    public string $endpoint;
    protected string $privateKeyId;
    protected string $privateKey;
    protected string $issuerId;
    public string $bundleId;

    protected Client $guzzleHttpClient;

    /**
     * @param string $endpoint     Web Service Endpoint
     * @param string $privateKeyId private key id
     * @param string $privateKey   private key
     * @param string $issuerId     IssuerID
     * @link https://developer.apple.com/documentation/appstoreserverapi/creating_api_keys_to_use_with_the_app_store_server_api private and issuerid see
     */
    public function __construct(string $endpoint, string $privateKeyId, string $privateKey, string $issuerId, $bundleId = null) {
        $this->endpoint = $endpoint;
        $this->privateKeyId = $privateKeyId;
        $this->privateKey = $privateKey;
        $this->issuerId = $issuerId;
        $this->bundleId = $bundleId;
        $this->guzzleHttpClient = GuzzleClient::create();
    }

    protected function jwt($bundleId = null): string {
        $payload = [
            'iss' => $this->issuerId,
            "iat" => time() - 10,
            "exp" => time() + 3590,
            "aud" => "appstoreconnect-v1",
            "nonce" => $this->uuid(),
            "bid" => $bundleId ?: $this->bundleId,
        ];
        return JWT::encode($payload, $this->privateKey, 'ES256', $this->privateKeyId);
    }

    protected function uuid(): string {
        $chars = md5(uniqid(mt_rand(), true));
        return substr($chars, 0, 8) . '-'
            . substr($chars, 8, 4) . '-'
            . substr($chars, 12, 4) . '-'
            . substr($chars, 16, 4) . '-'
            . substr($chars, 20, 12);
    }

    /**
     * @throws GuzzleException
     */
    protected function get($path, $bundleId = null): string {
        return $this->guzzleHttpClient->get($this->endpoint . $path, [
            'headers' => [
                'Authorization' => "Bearer " . $this->jwt($bundleId)
            ]
        ])->getBody()->getContents();
    }

    /**
     * @throws GuzzleException
     */
    protected function put($path, $body, $bundleId = null): string {
        return $this->guzzleHttpClient->put($this->endpoint . $path, [
            'headers' => [
                'Authorization' => "Bearer " . $this->jwt($bundleId),
                'Content-Type' => 'application/json'
            ],
            'body' => $body
        ])->getBody()->getContents();
    }

    /**
     * @throws GuzzleException
     */
    protected function post($path, $body, $bundleId = null): string {
        return $this->guzzleHttpClient->post($this->endpoint . $path, [
            'headers' => [
                'Authorization' => "Bearer " . $this->jwt($bundleId),
            ],
            'json' => $body,
        ])->getBody()->getContents();
    }

    /**
     * 查询内购记录
     * Get Transaction History
     * Get a customer’s in-app purchase transaction history for your app.
     * @param string|TransactionHistoryRequest $params
     * @throws GuzzleException
     */
    public function history(string $originalTransactionId, $params = '', $bundleId = null) {
        $path = '/inApps/v1/history/' . $originalTransactionId;

        if ($params instanceof TransactionHistoryRequest) {
            $path .= '?' . http_build_query($params->toArray());
        } elseif (!empty($params)) {
            $path .= '?revision=' . $params;
        }

        return HistoryResponse::make(json_decode($this->get($path, $bundleId), true));
    }

    /**
     * 获取所有订阅状态
     * Get All Subscription Statuses
     * Get the statuses for all of a customer’s auto-renewable subscriptions in your app.
     * @throws GuzzleException
     */
    public function subscriptions(string $originalTransactionId, $bundleId = null) {
        $path = '/inApps/v1/subscriptions/' . $originalTransactionId;

        return StatusResponse::make(json_decode($this->get($path, $bundleId), true));
    }

    /**
     * Send Consumption Information
     * Send consumption information about a consumable in-app purchase to the App Store after your server receives a consumption request notification.
     * @return void
     * @throws GuzzleException
     */
    public function transactionsConsumption(string $originalTransactionId, $requestBody, $bundleId = null) {
        $path = '/inApps/v1/transactions/consumption/' . $originalTransactionId;

        if ($requestBody instanceof ConsumptionRequest) {
            $requestBody = $requestBody->toArray();
        }

        $this->put($path, $requestBody, $bundleId);
    }

    /**
     * 使用订单ID查询内购
     * Look Up Order ID
     * Get a customer’s in-app purchases from a receipt using the order ID.
     * @return OrderLookupResponse ["status"=>0,"signedTransactions"=>[JWS]]
     * @throws GuzzleException
     */
    public function lookup(string $orderId, $bundleId = null) {
        $path = '/inApps/v1/lookup/' . $orderId;

        return OrderLookupResponse::make(json_decode($this->get($path, $bundleId), true));
    }

    /**
     * 获取退款历史
     * Get Refund History
     * Get a list of up to 50 of a customer’s refunded in-app purchases for your app.
     * @throws GuzzleException
     */
    public function refundLookup(string $originalTransactionId, $bundleId = null) {
        $path = '/inApps/v1/refund/lookup/' . $originalTransactionId;

        return RefundLookupResponse::make(json_decode($this->get($path, $bundleId), true));
    }

    /**
     * 延长订阅续订日期
     * Extend a Subscription Renewal Date
     * Extend the renewal date of a customer’s active subscription using the original transaction identifier.
     * @param ExtendRenewalDateRequest|array $requestBody
     * @return ExtendRenewalDateResponse
     * @throws GuzzleException
     */
    public function subscriptionsExtend(string $originalTransactionId, $requestBody, $bundleId = null) {
        $path = '/inApps/v1/subscriptions/extend/' . $originalTransactionId;

        if ($requestBody instanceof ExtendRenewalDateRequest) {
            $requestBody = $requestBody->toArray();
        }
        return ExtendRenewalDateResponse::make(json_decode($this->put($path, $requestBody, $bundleId), true));
    }

    /**
     * 通知历史
     * Get Notification History
     * Get a list of notifications that the App Store server attempted to send to your server.
     * @param NotificationHistoryRequest|array $requestBody
     * @return NotificationHistoryResponse
     * @throws GuzzleException
     */
    public function notificationHistory($requestBody, string $paginationToken = '', $bundleId = null) {
        $path = '/inApps/v1/notifications/history';
        if (!empty($paginationToken)) {
            $path .= '?paginationToken=' . $paginationToken;
        }
        if ($requestBody instanceof NotificationHistoryRequest) {
            $requestBody = $requestBody->toArray();
        }
        return NotificationHistoryResponse::make(json_decode($this->post($path, $requestBody, $bundleId), true));
    }

    /**
     * 请求测试通知
     * Request a Test Notification
     * Ask App Store Server Notifications to send a test notification to your server.
     * @return SendTestNotificationResponse
     * @throws GuzzleException
     */
    public function notificationsTest($bundleId = null) {
        $path = '/inApps/v1/notifications/test';
        return SendTestNotificationResponse::make(json_decode($this->post($path, [], $bundleId), true));
    }

    /**
     * Get Test Notification Status
     * Check the status of the test App Store server notification sent to your server.
     * @return CheckTestNotificationResponse
     * @throws GuzzleException
     */
    public function notificationsTestStatus(string $testNotificationToken, $bundleId = null) {
        $path = '/inApps/v1/notifications/test/' . $testNotificationToken;
        return CheckTestNotificationResponse::make(json_decode($this->get($path, $bundleId), true));
    }

}